#pragma once
#include<Windows.h>
#include<vector>
#include<thread>
#include<string>
#include<functional>

namespace SmakCEngineV2
{
	int Round(float f);
	std::wstring toWString(std::string str);
	std::string toString(std::wstring str);
	void ClearInputBuffer();
	inline short GetAsyncKeyStateF(int vKey)
	{
		if (GetConsoleWindow() != GetForegroundWindow())
			return 0;
		return GetAsyncKeyState(vKey);
	}

	// If a capital mode is used, CaseSensitive is implied
	// Use the filter constructor for more control
	struct InputStyleInfo
	{
		std::wstring opener;
		std::wstring closer;
		std::string c_opener;
		std::string c_closer;
	};
	enum class InputStyle
	{
		Normal,
		NoSpace,
		None
	};
	enum class Modes : char
	{
		Any,
		Number,
		Letters,
		CLetters, // Capital letters
		Hex,
		CHex, // Capital Hex
		IP,
		Custom
	};
	struct Rect
	{
		int left = 0;
		int top = 0;
		int right = 0;
		int bottom = 0;

		Rect() = default;
		Rect(int left, int top, int right, int bottom)
			: left(left), top(top), right(right), bottom(bottom)
		{
		}

		bool operator==(const Rect& other) const
		{
			return
				left == other.left &&
				top == other.top &&
				right == other.right &&
				bottom == other.bottom;
		}

	};

	struct Vector2
	{
		float x;
		float y;
		Vector2();
		Vector2(float X, float Y);
		bool IsNegative();
		//Convertion to string
		operator std::string() { return std::to_string(this->x) + ", " + std::to_string(this->y); };
		bool operator==(const Vector2& other) const;
	};

	class Object;
	class Layer;

	class VirtualObject
	{
	public:
		virtual ~VirtualObject() {};
		virtual void Update() = 0;
	};

	enum class Alignment
	{
		Center = 0b0000,
		Left = 0b1000,
		TopLeft = 0b1100,
		Top = 0b0100,
		TopRight = 0b0110,
		Right = 0b0010,
		BottomRight = 0b0011,
		Bottom = 0b0001,
		BottomLeft = 0b1001
	};
	class Object : public VirtualObject
	{
	private:
		std::wstring Sprite;
		std::string C_Sprite;
		std::string BC_Sprite;
		std::string S_Sprite;
		std::string Attr_Sprite;
		int height;
		int width;
		bool destroyed;
		Layer* layer;
		std::vector<Object*> children;
		Object* Parent;
		Vector2 GetGlobalPosRec();
		Rect GetBounds();
	protected:
		std::function<void(Object*)> func;
		~Object();
		friend class Layer;
	public:
		bool render;
		std::string type;
		std::string name;
		Vector2 position;
		Alignment alignment;
		bool useBounds;
		Rect padding;
		Object(Vector2 pos, std::wstring sprite);
		Object(Vector2 pos, std::string sprite) : Object(pos, toWString(sprite)) {}
		Object(float x, float y, std::wstring sprite) : Object(Vector2(x, y), sprite) {}
		Object(float x, float y, std::string sprite) : Object(Vector2(x, y), toWString(sprite)) {}
		void Destroy();
		bool Destroyed();
		//Sprite
		void SetSprite(std::wstring sprite);
		void SetSprite(std::string sprite) { SetSprite(toWString(sprite)); };
		std::wstring GetSprite();
		//Color Sprite
		void SetColorSprite(std::string sprite);
		void MonoColor(char color);
		std::string GetColorSprite();
		//Background Color Sprite
		void SetBColorSprite(std::string sprite);
		void BMonoColor(char color);
		std::string GetBColorSprite();
		//Solid Sprite (# for solid)
		void SetSolidSprite(std::string sprite);
		void MonoSolid();
		std::string GetSolidSprite();
		//Attribute Sprite
		void SetAttrSprite(std::string sprite);
		void MonoAttr(char attr);
		std::string GetAttrSprite();

		void OnUpdate(void (*Funk)(Object*));
		void Update();
		int GetHeight();
		int GetWidth();
		bool IsOutOfBounds();
		Vector2 GetCenter(bool global = false);
		float GetRenderedTop(bool global = false);
		float GetTop(bool global = false);
		float GetRenderedBottom(bool global = false);
		float GetBottom(bool global = false);
		float GetRenderedLeft(bool global = false);
		float GetLeft(bool global = false);
		float GetRenderedRight(bool global = false);
		float GetRight(bool global = false);

		Vector2 GetGlobalPos();
		Object* GetParent();
		std::vector<Object*> GetChildren();
		void AddChild(Object* Object);
		void RemoveChild(Object* Object);
		void DeleteChild(Object* object);
		bool HasChild(Object* Object);

		bool Clicked();
		bool IsMouseOver();

		bool Collides(Object* obj);

	};
	class TextInput : public Object
	{
		void RefreshText();
		int MaxSize;
		bool CaseSensitive;
		uint64_t time = 0;
		char colorThing = 'F';
		std::wstring opener = L"";
		std::string copener = "";
		std::wstring closer = L"";
		std::string ccloser = "";
	protected:
		~TextInput();
		bool Focused;
	public:
		std::function<void(Object*)> OnEnter;
		Modes Mode;
		int MinWidth;
		int CursorIndex;
		bool insertMode;
		std::wstring Text;
		std::wstring Placeholder;
		std::wstring Filter = L"";
		TextInput(Vector2 position, std::wstring placeholder, int MaxSize, Modes Tm = Modes::Any, InputStyle style = InputStyle::Normal, InputStyleInfo* info = nullptr);
		TextInput(Vector2 position, std::wstring placeholder, int maxSize, std::wstring Filter, bool cagseSensitive = false, InputStyle style = InputStyle::Normal, InputStyleInfo* info = nullptr);
		void Update();
		void Focus();
		bool IsFocused();
	};
	class Slider : public Object
	{
		void Refresh();
	protected:
		~Slider();
	public:
		bool Focused;
		char Empty = '-';
		char Filled = '|';
		char C_Empty = '7';
		char C_Filled = 'F';
		InputStyleInfo style;
		int MinVal, MaxVal, Value;
		Slider(Vector2 pos, int minVal, int maxVal, int value, InputStyle style = InputStyle::Normal, InputStyleInfo* info = nullptr);
		void Update();
		void Focus();
		bool IsFocused();
	};
	class Button : public Object
	{
		std::function<void(Object*)> click = 0;
		void ClickCheck();
	public:
		Button(Vector2 pos, std::wstring text, InputStyle style = InputStyle::Normal, InputStyleInfo* info = nullptr);
		void OnClick(std::function<void(Object*)> click);
		void Update();
	};

	class Layer
	{
		std::vector<Object*> Objects;
	public:
		std::string name;
		Layer(std::string name);
		std::vector<Object*> GetObjects();
		bool Contains(Object* Object);
		void Add(Object* Object);
		void Remove(Object* Object);
		void Delete(Object* Object);
	};

	class Renderer
	{
		bool SetupConsole();
		void RenderObject(Object* object, CHAR_INFO* screen, bool* screenMask, int screenWidth, int screenHeight, Rect boundry);
		void RenderLoop(void);
		std::vector<Layer*> Layers;
		int FPS;
		bool Focused;
		int Width;
		int Height;
	public:
		std::function<void(Vector2)> ResizeCallback;
		uint64_t DeltaTime;
		HANDLE hInput;
		HANDLE hOutput;
		Vector2 CameraPos;
		std::thread RenderThread;
		Layer* Objects;
		bool RenderOffFocus;
		bool Active;
		int FrameDelayMs;
		Renderer();
		Renderer(bool RenderOffFocus);
		int GetWidth();
		int GetHeight();
		Vector2 GetCenter();
		Layer* GetLayer(std::string name);
		void AddLayer(Layer* layer);
		COLORREF RGBToColorRef(uint8_t R, uint8_t G, uint8_t B);
		void GetColorPallete(COLORREF* colors);
		void SetColorInPallete(int index, COLORREF color);
		void SetColorPallete(COLORREF* colors);

		bool InFocus();
		int GetFPS();
	};
	extern Renderer* MainRenderer;

	extern Layer* Menus;

	class MenuObject : public Object
	{
	public:
		bool CloseIfOutside = true;
		MenuObject();
		MenuObject(Vector2 pos, int width, int height);
		void Resize(int width, int height);
	};

	class ContextChoice
	{
	public:
		std::function<void(Object*)> Function;
		bool submenu;
		std::wstring name;
		ContextChoice(std::wstring Name, void(*Func)(Object*), bool Submenu);
	};

	class ContextMenu : public MenuObject
	{
	public:
		ContextMenu(std::vector<ContextChoice> choices, Vector2 pos);
	};
	std::wstring StringToSprite(std::wstring str, int w);
	int StringToSpriteWidth(std::wstring str);
	void CloseMenu();
	void CloseMenuO(Object* obj);
	void CloseAllMenus();
	void Alert(std::wstring Title, std::wstring info);
	template<typename ResT>
	void POClick(Object* o);
	template<typename ResT>
	class PromptOptionObject : public Button
	{
	public:
		ResT Value;
		ResT** RetVal;
		PromptOptionObject(Vector2 pos, std::wstring text, ResT** ret, ResT Value) : Button(pos, text)
		{
			OnClick(POClick<ResT>);
			this->RetVal = ret;
			this->Value = Value;
		}
	};

	template<typename ResT>
	void POClick(Object* o)
	{
		(*((PromptOptionObject<ResT>*)o)->RetVal) = &((PromptOptionObject<ResT>*)o)->Value;
		CloseMenu();
	}

	template<typename ResT>
	struct PromptOption
	{
		std::wstring Name;
		ResT Value;
		PromptOption(std::wstring Name, ResT Value)
		{
			this->Name = Name;
			this->Value = Value;
		}
	};
	template<typename ResT>
	ResT Prompt(std::wstring Title, std::wstring info, std::vector<PromptOption<ResT>> options)
	{
		if (options.size() < 1)
			throw "You need at least one option in a prompt!";

		int width = StringToSpriteWidth(info);
		if (width > MainRenderer->GetWidth() - 7)
			width = MainRenderer->GetWidth() - 7;
		if (width < 33)
			width = 33;

		MenuObject* menu = new MenuObject(MainRenderer->GetCenter(), width + 5, 3);
		Menus->Add(menu);
		menu->CloseIfOutside = false;
		Object* title = new Object(0, 0, Title);

		menu->AddChild(title);
		Object* infoObj = new Object(0, 0, StringToSprite(info, width));

		menu->AddChild(infoObj);
		ResT* Value = 0;

		menu->Resize(width+5, infoObj->GetHeight() + options.size() + 1);

		title->alignment = Alignment::Left;
		title->position = Vector2(
			-(menu->GetWidth() / 2.0f) + 2.5f,
			(menu->GetHeight() / 2 - 0.5f) - 1
		);
		infoObj->alignment = Alignment::TopLeft;
		infoObj->position = Vector2(
			-(menu->GetWidth() / 2.0f) + 5.5f,
			(menu->GetHeight() / 2) - 2
		);

		float YStart = ((menu->GetHeight() / 2.0f)-0.5f) - 1 - 1 - infoObj->GetHeight();
		for (int i = 0; i < options.size(); i++)
		{
			PromptOptionObject<ResT>* button = new PromptOptionObject<ResT>(Vector2(0, 0), options[i].Name, &Value, options[i].Value);
			button->position = Vector2(
				(menu->GetWidth() / 2) - (button->GetWidth() / 2) - 2.5f,
				YStart - i
			);
			menu->AddChild(button);
		}
		while (Value == nullptr)
		{
			Sleep(10);
		}

		return *Value;
	}

	class Mouse
	{
		struct KeyState
		{
			bool Pressed;
			bool Released;
			bool Held;
		};
		static bool NewState[5];
		static bool OldState[5];
	public:
		static Vector2 Position;
		static int VScroll;
		static KeyState Keys[5];

		static void Update(HANDLE hInput, Vector2 mod);
	};
	enum KeyboardMod : unsigned short
	{
		NONE = 0,
		LCTRL = 0x08,
		RCTRL = 0x04,
		CTRL = 0x1000,

		LALT = 0x02,
		RALT = 0x01,
		ALT = 0x2000,

		SHIFT = 0x10,
		//CAPSLOCK = 0x80,
		//NUMLOCK = 0x20,
		//SCROLLLOCK = 0x40,
	};
	class Keyboard
	{
		static bool states[0x100];
		static std::vector<KEY_EVENT_RECORD> records;
	public:
		static void Update(KEY_EVENT_RECORD rec);
		static bool GetKey(unsigned char vKey, bool held = false, KeyboardMod mods = NONE);
		static std::vector<KEY_EVENT_RECORD> GetAllChars();
		static void Clear();
	};
}